# =====================================================
#      PAS_3d_analysis.py
#   analysis of the data to get all PAS features
#   as odd-even shift, ghost counts, etc
#   Versions
#   2020-04-19 V.1.0, Fedorov, initial
#   2020-04-20 V.1.1, Fedorov, odd/even E steps elevations
#   2020-04-20 V.2.0, Fedorov, correction og ghost count
#   2020-05-05 V.2.1, Fedorov, 
#   2020-05-10 V.3.0, Fedorov, all calibrations in pas_process.py package, Full calibation tables, Odd-Even plots
#   2020-05-13 V.3.1, Fedorov, new Azimuth plot and 2D distributions
#   2020-05-14 V.3.2, Fedorov, elevation correction is in calibration tables
#   2020-05-14 V.3.3, Fedorov, VDF reading also
# ====================================================

# ----------------------------------------------------
#  PLPLOT and other staf setup
# ----------------------------------------------------
import sys
sys.path.insert (0, "/usr/share/plplot5.14.0/examples/python/")

from PIL import Image
import math
import numpy

from plplot_python_start import *
import plplot as plot

# ----------------------------------------------------
# CDF reader
# ---------------------------------------------------
import spacepy
import datetime
import time
from spacepy import pycdf

#------------------------------------------------------
# PAS package
#------------------------------------------------------
import pas_process as pas

#-----------------------------------------------------
# Debugging
#-----------------------------------------------------
# import pdb
# pdb.set_trace()

# --------------------------------------------------------
# Functions responsable for the different axis frame
# preparation
# --------------------------------------------------------
def eAxis():
    global wxmin, wxmax, wymin, wymax
    wxmin = numpy.log10(eRange[0])
    wxmax = numpy.log10(eRange[1])
    wymin = numpy.log10(cntRange[0])
    wymax = numpy.log10(cntRange[1])
    plot.plwind(wxmin, wxmax, wymin, wymax)
    plot.plbox("bclnst", 0.0, 0, "bclnstv", 0.0, 0) # both with axis
    plot.pllab( "Energy, eV", "count", "" )         # axis titles

def vAxis():
    global wxmin, wxmax, wymin, wymax
    wxmin = numpy.log10(vRange[0])
    wxmax = numpy.log10(vRange[1])
    wymin = numpy.log10(vdfRange[0])
    wymax = numpy.log10(vdfRange[1])
    plot.plwind(wxmin, wxmax, wymin, wymax)
    plot.plbox("bclnst", 0.0, 0, "bclnstv", 0.0, 0) # both with axis
    plot.pllab( "Vr, km/s", "VDF, cm-6 s+3", "" )   # axis titles
    
def elAxis():    
    global wxmin, wxmax, wymin, wymax
    wxmin = elRange[0]
    wxmax = elRange[1]
    wymin = numpy.log10(cntRange[0])
    wymax = numpy.log10(cntRange[1])
    plot.plwind(wxmin, wxmax, wymin, wymax)
    plot.plbox("bcnst", 0.0, 0, "bclnstv", 0.0, 0) # both with axis
    plot.pllab("Elevation, deg", "count", "" )    # axis titles
    
def azAxis():
    global wxmin, wxmax, wymin, wymax
    wxmin = azRange[0]
    wxmax = azRange[1]
    wymin = numpy.log10(cntRange[0])
    wymax = numpy.log10(cntRange[1])
    plot.plwind(wxmin, wxmax, wymin, wymax)
    plot.plbox("bcnst", 0.0, 0, "bclnstv", 0.0, 0)  # both with axis
    plot.pllab("Azimuth, deg", "count", "" )        # axis titles
# ---------------------------------------------------    

# -----------------------------------------------------
# Data Intefaces
# -----------------------------------------------------
DEV = "SVG"
# DEV = "X"

toPlot = "E"
# toPlot = "AZ"
# toPlot = "EL"
# toPlot = "RVDF"
switch = {'E': eAxis, 'EL': elAxis, 'AZ': azAxis, 'RVDF': vAxis}

dDir = "../DATA_L1/"
# normalFile = "20200428/solo_L1_swa-pas-3dn_20200428_V00.cdf"
#n ormalFile = "20200415/solo_L1_swa-pas-3d_20200415_V00.cdf"
# normalFile = "20200505/solo_L1_swa-pas-3d_20200505_V00.cdf"
# normalFile = "20200508/solo_L1_swa-pas-3d_20200508_V00.cdf" # recordN = 1000
normalFile = "20200511/solo_L1_swa-pas-3d_20200511_V00.cdf"   # recordN = 1000
outSVG = 'xxxx.svg'

# --------------------------------------
# Record number
# --------------------------------------
# recordN = 52 # Normal wind a little hot (2020-04-28T11:03:24)
# recordN = 304 # Normal in Elevation but very hot (2020-04-28T11:26:44)
# recordN = 508 # Very deviated in Elevation (2020-04-28T11:40:56)
# recordN = 1461 # double paeak on 14 Apr 2020
recordN = 2452 # behaviour during calibration 5 May 2020

# -----------------------------------------------------
# Constants and ranges
# ----------------------------------------------------
eRange =  [100.0, 10000.0]
elRange = [ -30.0, 30.0 ]
azRange = [ -26.0, 43.0 ]
cntRange = [1.0, 5000.0]
vRange = [100.0, 1000.0]
vdfRange = [1.0e-23, 1.0e-18]
elStep = 1.5  # deg for the HR plots
azStep = 1.5  # deg

#----------------------------------------------------
# init the PL PLOT
# ---------------------------------------------------
if(DEV == "X"):
    plot.plsdev('xwin')
    xp1 = 20
    yp1 = 20
    xleng1 = 600
    yleng1 = 600
    xoff1 = 0
    yoff1 = 0
    axthick = 2
    lthick = 2

if(DEV == "SVG"):
    plot.plsdev('svg')
    plot.plsfnam(outSVG)
    xp1 = 10
    yp1 = 10
    xleng1 = 400
    yleng1 = 400
    xoff1 = 0
    yoff1 = 0
    axthick = 1.0
    lthick = 2.0

plot.plscolbg(255,255,255)  # white background
plot.plscol0(15, 0, 0, 0)        # make black in #15
plot.plspage(xp1, yp1, xleng1, yleng1, xoff1, yoff1)
plot.plinit ()

# ----------------------------------------------------
# Plot the frame corresponding to toPlot value
# ----------------------------------------------------
plot.pladv(0)
plot.plcol0(15)                      # black color for axis
plot.plwidth(axthick)               # thick of axis

plot.plvpor(0.15, 0.85, 0.1, 0.9)    # vewport. Points correcpond to the coordinates in window
switch[toPlot]()                     # make correspondig axis
# ----------------------------------------------------

#-------------------------------------------------
# Read Calibration data
#-------------------------------------------------
eArrMin = pas.eTable["emin"]
eArrMax = pas.eTable["emax"]
eArr = pas.eTable["ecent"]
vArr = pas.eTable["vcent"]*1.0e-5

elArrMin = pas.elTable["elmin"]
elArrMax = pas.elTable["elmax"]
elArr = pas.elTable["elcent"]

azArrMin = pas.azTable['azmin']
azArrMax = pas.azTable['azmax']
azArr = pas.azTable['azcent']

elCorrArr = pas.elCorrTab["deg"]

#------------------------------------------
# Read data file
# -----------------------------------------
normalCDF = pycdf.CDF(dDir+normalFile)
epoch = normalCDF['Epoch'][recordN]
spectrum3D = normalCDF['COUNTS'][recordN,:,:,:]
sE = normalCDF['FIRST_ENERGY'][recordN]
sEl = normalCDF['FIRST_ELEVATION'][recordN]
nE = normalCDF['NB_ENERGY'][recordN]
nEl = normalCDF['NB_ELEVATION'][recordN]
print("Epoch,sE,nE, sEl, Nel: ",epoch,sE,nE,sEl,nEl)
# ---------------------------------------
# cut the full array and transfer it from list of integers
# to the numpy.array of floats
# --------------------------------------
spectrum3D = numpy.array(spectrum3D,dtype = float)
spectrum3Dnul = numpy.zeros(spectrum3D.shape,dtype = float)
# ---------------------------------------------
# Set Nulls out of boundaries
# ---------------------------------------------
for iel in range(nEl):
    for ie in range(nE):
        spectrum3Dnul[:,iel+sEl,ie+sE] =  spectrum3D[:,iel+sEl,ie+sE]
#------------------------------------------
# Clean the Spectrum
#------------------------------------------
spectrum3Dclean = pas.specCorrection(spectrum3Dnul, 0, 9, 0, 96)

#------------------------------------------
# Procession of Energy spectrum
#-------------------------------------------
# eSpectrum = numpy.reshape(spectrum3Dclean[5,5,:], 96)
eSpectrum = numpy.sum(numpy.sum(spectrum3Dclean,axis = 0), axis = 0)
# -----------------------------------------
# Get maximum step in Energy
#- -----------------------------------------
index = numpy.argmax(eSpectrum)
print("Energy max and 2 around",index, eSpectrum[[index-1,index,index+1]])

# -----------------------------------------
# Indices of even and odd energies
#------------------------------------------
evArr = numpy.arange(sE,sE+nE,2,dtype=int)
oddArr = evArr+1
# -------------------------------------------
# min - max in energy for plot purposes
# -------------------------------------------
leArrFull = numpy.zeros(192)
leArrFull[0:192:2] = eArrMax
leArrFull[1:192:2] = eArrMin
# ---------------------------------------------------------------

# ---------------------------------------------------------------
# Velocity distribution function in Vr (jast integration over Az and El
# ---------------------------------------------------------------
vdf = pas.calcVDF(spectrum3D, sE, sEl, nE, nEl)
rvdf = numpy.sum(numpy.sum(vdf,axis = 0), axis = 0)
# pdb.set_trace()

# ---------------------------------------------
# Elevation summary HR spectrum
# elStep  deg step inside the elRange
# ---------------------------------------------
elSize = int((elRange[1] - elRange[0])/elStep+0.5)
elArrPlot = numpy.zeros(elSize*2,dtype=float)
elArrF = numpy.zeros(elSize,dtype=float)       # mid points
elSpectrum = numpy.zeros(elSize,dtype=float)   # array to plot

for iel in range(elSize):
    point = elRange[0]+ float(iel)*elStep + elStep/2.0
    elArrPlot[iel*2] = point
    elArrPlot[iel*2+1] = point
    elArrF[iel] = point
elSpectrumEven =  numpy.zeros(elSize,dtype=float)
elSpectrumOdd =  numpy.zeros(elSize,dtype=float)

# --------------------------------------
# sum in azimuth and even/odd energies
# --------------------------------------
for iaz in range(11):
    for iel in range(9):
        for iell in range(elSize):
            for iee in range(0,96,2):
                if ((elArrMin[iaz,iel] < elArrF[iell]+elCorrArr[iee]) and (elArrMax[iaz,iel] >= elArrF[iell]+elCorrArr[iee])):  # inside for evens
                    elSpectrumEven[iell] += spectrum3Dclean[iaz,iel,iee]
                if ((elArrMin[iaz,iel] < elArrF[iell]+elCorrArr[iee+1]) and (elArrMax[iaz,iel] >= elArrF[iell]+elCorrArr[iee+1])):  # inside for odds
                    elSpectrumOdd[iell] += spectrum3Dclean[iaz,iel,iee+1]
elSpectrum = elSpectrumEven + elSpectrumOdd
# ---------------------------------------------

# ----------------------------------------
# Azimuth HR spectrum (the same princip as for Elevation)
# ----------------------------------------
azSize = int((azRange[1] - azRange[0])/azStep+0.5)
azArrPlot = numpy.zeros(azSize*2,dtype=float)  # to plot with min-max values
azArrF = numpy.zeros(azSize,dtype=float)       # mid points
azSpectrum = numpy.zeros(azSize,dtype=float)   # array to plot

for iaz in range(azSize):
    point = azRange[0]+ float(iaz)*azStep + azStep/2.0
    azArrPlot[iaz*2] = point
    azArrPlot[iaz*2+1] = point
    azArrF[iaz] = point
azSpectrumEven =  numpy.zeros(azSize,dtype=float)
azSpectrumOdd =  numpy.zeros(azSize,dtype=float)

# --------------------------------------
# sum in elevation and even/odd energies
# --------------------------------------
evArr = numpy.array([58])
oddArr = numpy.array([59])
for iel in range(4,5):
    for iaz in range(11):
        for iazl in range(azSize):
            if ((azArrMin[iaz,iel] < azArrF[iazl]) and (azArrMax[iaz,iel] >= azArrF[iazl])):  # inside for evens
                azSpectrumEven[iazl] += numpy.sum(spectrum3Dclean[iaz,iel,evArr])
            if ((azArrMin[iaz,iel] < azArrF[iazl]) and (azArrMax[iaz,iel] >= azArrF[iazl])):  # inside for odds
                azSpectrumOdd[iazl] += numpy.sum(spectrum3Dclean[iaz,iel,oddArr])
azSpectrum = azSpectrumEven + azSpectrumOdd

if toPlot == 'E':
    mimaeSpectrum = numpy.zeros((nE*2), dtype=float)
    for ie in range(nE):       # special spectrum to plot
        mimaeSpectrum[2*ie] = eSpectrum[ie+sE] + cntRange[0]
        mimaeSpectrum[2*ie+1] = eSpectrum[ie+sE] + cntRange[0]
    leArr = leArrFull[2*sE:2*(sE+nE)]
    plot.plcol0(9)        # Blue
    plot.plwidth(lthick)  # line width
    plot.plline(numpy.log10(leArr), numpy.log10(mimaeSpectrum))

    plot.plcol0(15)       # BLACK
    plot.plptex( wxmax*0.52 , wymax*0.95 , 0 , 0 , 0 , str(epoch)+" UTC");

if toPlot == 'RVDF':
    mimaeSpectrum = numpy.zeros(192, dtype=float)
    lvArr = numpy.zeros(192, dtype=float)
    for ie in range(96):       # special spectrum to plot
        mimaeSpectrum[2*ie] = rvdf[ie] + vdfRange[0]
        mimaeSpectrum[2*ie+1] = rvdf[ie] + vdfRange[0]
        lvArr[2*ie] = vArr[ie]
        lvArr[2*ie+1] = vArr[ie]
        
    plot.plcol0(9)        # Blue
    plot.plwidth(lthick)  # line width
    plot.plline(numpy.log10(lvArr), numpy.log10(mimaeSpectrum))
    
    plot.plcol0(15)       # BLACK
    plot.plptex( wxmin*1.0 , wymax*0.95 , 0 , 0 , 0 , str(epoch)+" UTC");

# -----------------------------------------
# make and plot elevation spectrum
#------------------------------------------
if toPlot == 'EL':        
    plot.plwidth(lthick)  # line width
    plot.plcol0(9)        # Blue for even
    plot.plline(elArrF,numpy.log10(elSpectrumEven+cntRange[0]))
    plot.plcol0(1)        # Red for odd
    plot.plline(elArrF,numpy.log10(elSpectrumOdd+cntRange[0]))
    plot.plcol0(15)       # BLACK
    plot.plline(elArrF,numpy.log10(elSpectrum+cntRange[0]))
    
    plot.plcol0(15)       # BLACK
    plot.plptex( wxmin*0.8 , wymax*0.95 , 0 , 0 , 0 , str(epoch)+" UTC");
 
if toPlot == 'AZ':   
    plot.plwidth(lthick)  # line width
    plot.plcol0(9)        # Blue for even
    plot.plline(azArrF,numpy.log10(azSpectrumEven+cntRange[0]))
    plot.plcol0(1)        # Red for odd
    plot.plline(azArrF,numpy.log10(azSpectrumOdd+cntRange[0]))
    plot.plcol0(15)       # BLACK
    plot.plline(azArrF,numpy.log10(azSpectrum+cntRange[0]))
    
    plot.plcol0(15)       # BLACK
    plot.plptex( wxmin*0.8 , wymax*0.95 , 0 , 0 , 0 , str(epoch)+" UTC");
 
# -----------------------------------------------------
# End of Plot, close all
# -----------------------------------------------------
plot.plend()



